import os
import sys
import inspect
from dataclasses import is_dataclass
from collections import OrderedDict, defaultdict, deque, Counter, ChainMap, UserDict, UserList, UserString
from importlib.util import find_spec

stdlib_paths = {
  os.path.normpath(path) for path in sys.path
  if os.path.isdir(path) and (
    "lib/python" in path and "site-packages" not in path and "dist-packages" not in path
  )
}

def is_c_wrapper(obj):
  """
  Check if the object is a C wrapper.
  Objects in this type are non-interesting, and don't need to futher checking its attributes.

  method-wrapper: methods implemented in C code, bound to an instance of a class.
  builtin_function_or_method: built-in functions or methods implemented in C. Not bound to a class instance and can be called directly.
  """
  try:
    if type(obj).__name__ in ["method-wrapper", "builtin_function_or_method", "method_descriptor", "getset_descriptor"]:
      return True
  except AttributeError:
    pass
  return False

def is_from_standard_library(obj):
  module = inspect.getmodule(obj)
  if module is None:
    return is_c_wrapper(obj)

  module_file = getattr(module, "__file__", None)
  # No module file, then it's module builtins
  if module_file is None:
    return True
  
  module_file = os.path.normpath(module_file)
  for path in stdlib_paths:
    if module_file.startswith(os.path.join(path, "")):
      return True
  return False

def is_standard_module(module):
  """
  Check if the module is part of the Python standard library.

  @params module: The module to check.
  @return: True if the module is part of the standard library, False otherwise.
  """
  if module is None:
    return False
  
  module_file = getattr(module, "__file__", None)
  if module_file is None:
    return False
  
  module_file = os.path.normpath(module_file)
  for path in stdlib_paths:
    if module_file.startswith(os.path.join(path, "")):
      return True
  return False

def get_type(val):
  """
  Get the type information of the object, including the full module path for class instances.

  @params val: The object to get the type information for.
  @return: A string representing the type of the object.
            Returns "class" for classes,
            full module-qualified name (e.g., `module.submodule.ClassName`) for class instances,
            and type name for other objects.
  """
  if val is None:
    return "None"

  # If val has a class type, return "class"
  if isinstance(val, type):
    return "class"

  obj_type = type(val)

  try:
    obj_type_str = str(obj_type)
    if obj_type_str.startswith("<class '"):
      cleaned_type_string = obj_type_str.replace("<class '", "").replace("'>", "")
      return cleaned_type_string
  except AttributeError:
    pass

  return obj_type.__name__

def is_class(type_info):
  """
  Check if the type name represents a class.

  @params type_info: The type name to check.
  @return: True if the type name represents a class, False otherwise.
  """
  return type_info in {"class", "ABCMeta", "method_descriptor", "type"}
  
def is_callable(type_info):
  """
  Check if the type name represents a callable.

  @params type_info: The type name to check.
  @return: True if the type name represents a callable, False otherwise.
  """
  return type_info in {
    "function", "builtin_function_or_method",
    "slot wrapper", "method-wrapper"
  }

def is_instance(type_info):
  """
  Check if the type name represents an instance.

  @params type_info: The type name to check.
  @return: True if the type name represents an instance, False otherwise.
  """
  return '.' in type_info

def is_data_structure(type_info):
  """
  Check if the type name represents a data structure.

  @params type_info: The type name to check.
  @return: True if the type name represents a data structure, False otherwise.
  """
  return type_info in {
    "dict", "list", "tuple", "set", "OrderedDict",
    "defaultdict", "deque", "Counter", "ChainMap",
    "UserDict", "UserList", "UserString"
  }

def support_getField_op(type_info):
  """
  Check if the type supports `obj[1]` operation.

  @params type_info: The type name to check.
  @return: True if the type supports getField, False otherwise.
  """
  return type_info in {
    "list", "tuple", "dict", "str", "OrderedDict",
    "defaultdict", "deque", "UserList", "UserDict",
    "ChainMap", "Counter"
  }

def is_primitive(type_info):
  """
  Check if the type is non-primitive.

  @params type_info: The type name to check.
  @return: True if the type is non-primitive, False otherwise.
  """
  return type_info in {
    "int", "float", "complex", "bool", "str"
  }

def get_name_info(val):
  """
  Get appropriate name for different object types.

  @params val: The object to inspect.
  @return: Name for modules/callables, value for strings, type name otherwise.
  """
  if val is None:
    return "None"

  type_name = type(val).__name__

  if type_name == "method-wrapper":
    return getattr(val, "__name__", "unknown_method")
  elif type_name == "builtin_function_or_method":
    return getattr(val, "__name__", "unknown_method")
  elif isinstance(val, str):
    return val
  elif type_name == "module":
    return getattr(val, "__name__", "UnknownModule")
  elif type_name == "function":
    return getattr(val, "__name__", "unknown_function")
  elif type_name == "type":
    return getattr(val, "__name__", "UnknownClass")
  elif is_data_structure(type_name):
    return type_name
  else:
    return type_name

def is_c_written(func, type_info):
  """
  Check if the function is written in C.

  @params func: The function to check.
  """
  if not is_callable(type_info):
    return False
  
  try:
    if type(func).__flags__ & (1<<9):
      return True
    else:
      return False
  except Exception as e:
    return False


import sys
sys.path.insert(0, "/home/jackfromeast/Desktop/python-class-pollution/polluter")

# import math
import collections
import random
import re
from polluter import Pollutable
import polluter as pl


po = Pollutable(re, max_layer=-1, lookup_type="getAttr")
modules = po.select("type=module")
callables = po.select("type=callable")
dicts = po.select("type=dict")
classes = po.select("type=class")
strings = po.select("type=string")
objects = po.select("type=object")
sys = po.select("type=callable&name=open")
sys_like = po.select("type=callable&namelike=open")

# print("Modules:")
# for path, value in modules.items():
#   print(f"{path}: {value}")

# # Select callables
# print("\nCallables:")
# for path, value in callables.items():
#   print(f"{path}: {value}")

# print("\nDicts:")
# for path, value in dicts.items():
#   print(f"{path}: {value}")

# print("\nClasses:")
# for path, value in classes.items():
#   print(f"{path}: {value}")

# print("\nStrings:")
# for path, value in strings.items():
#   print(f"{path}: {value}")

# print("\nObjects:")
# for path, value in objects.items():
#   print(f"{path}: {value}")

print("\nSys:")
for path, value in sys.items():
  print(f"{path}: {value}")

print("\nSys-like:")
for path, value in sys_like.items():
  print(f"{path}: {value}")


#!/usr/bin/env python3
"""
@description
---------------------
This script contains the useful functions to find the pollutables reflectively during runtime.

import polluter as pl
po = new pl.Pollutable(obj, max_layer=1, lookup_type="getAttr")
"""
import inspect
import logging
import re
import types
import collections.abc
from .utils import get_type, get_name_info, is_instance, is_c_wrapper, is_from_standard_library, is_primitive, is_c_written, support_getField_op, is_callable

logging.basicConfig(
  level=logging.INFO,
  format='%(asctime)s - %(levelname)s - %(message)s'
)

class Pollutable:
  def __init__(self, target, max_layer=3, lookup_type="getAttr"):
    """
    @description
    ---------------------
    The class to represent the pollutable object.

    @params target: The object to pollute.
    @params max_layer: The maximum layer to search for pollutables.
    @params lookup_type: The type of pollutables to find. Default is "getAttr" and alternative is "getBoth".
    """ 
    self.target = target
    self.max_layer = max_layer
    self.lookup_type = lookup_type

    self.visited = set()
    self.summary = self.search(max_layer)
  
  def search(self, max_layer=-1):
    """
    @description
    ---------------------
    Find all the pollutables in the given object with logging support.

    @params max_layer: The maximum layer to search for pollutables.
    """
    if max_layer == -1:
      max_layer = self.max_layer

    logging.debug(f"Starting find_all_pollutables. Type: {self.lookup_type}, Layer: {0}, Max Layer: {max_layer}")

    def lookup_getattr(obj, name):
      try:
        ret = getattr(obj, name)
        logging.debug(f"Accessed attribute: {name}")
        return (ret, 'attr')
      except AttributeError as e:
        # logging.warning(f"Could not access attribute '{name}': {str(e)}")
        return None
      except TypeError as e:
        return None

    def lookup_getboth(obj, name):
      if hasattr(obj, '__getitem__'):
        try:
          ret = obj[name]
          logging.debug(f"Accessed item: {name}")
          return (ret, 'item')
        except (KeyError, IndexError, TypeError, AttributeError) as e:
          logging.debug(f"Failed to access item '{name}': {str(e)}")
          pass

      try:
        ret = getattr(obj, name)
        logging.debug(f"Accessed attribute: {name}")
        return (ret, 'attr')
      except AttributeError as e:
        # logging.debug(f"Could not access attribute or item '{name}': {str(e)}")
        return None
      
    if self.lookup_type == "getAttr":
      logging.info("Using 'getAttr' method to find pollutables.")
      self.summary = self.find_all_pollutables(self.target, 0, max_layer, False, lookup_getattr, '')
    elif self.lookup_type == "getBoth":
      logging.info("Using 'getBoth' method to find pollutables.")
      self.summary = self.find_all_pollutables(self.target, 0, max_layer, False, lookup_getboth, '')
    else:
      logging.error(f"Unknown type specified: {self.lookup_type}")
      raise Exception(f"Unknown type: {self.lookup_type}")

    return self.summary
  
  def find_all_pollutables(self, obj, layer=0, max_layer=1, callable_only=False, lookup_func=None, parent_full_path=''):
    """Find all pollutables in the given object using getattr recursively, logging module names (BFS version)."""
    import collections
    logging.debug(f"Starting BFS for pollutables. Type: {type(obj)}")

    pollutables = {}
    queue = collections.deque()
    queue.append((obj, layer, parent_full_path))

    while queue:
      current_obj, current_layer, current_path = queue.popleft()
      if current_layer >= max_layer and max_layer != -1:
        continue

      obj_id = id(current_obj)
      if obj_id in self.visited:
        continue
      self.visited.add(obj_id)

      attributes_to_be_checked = dir(current_obj) + ["__base__", "__bases__"]
      items_to_be_checked = []
      if isinstance(current_obj, collections.abc.Mapping):
        items_to_be_checked.extend(current_obj.keys())
      elif isinstance(current_obj, collections.abc.Sequence) and not isinstance(current_obj, (str, bytes, bytearray)):
        items_to_be_checked.extend(range(len(current_obj)))
      attributes_to_be_checked.extend(items_to_be_checked)

      for name in attributes_to_be_checked:
        if lookup_func:
          result = lookup_func(current_obj, name)
          if not result:
            continue
          value, access_type = result
        else:
          continue

        if callable_only and not callable(value):
          logging.debug(f"Skipping non-callable attribute: {name}")
          continue

        # Build current_full_path based on access_type and current_path
        if access_type == 'attr':
          current_full_path = f"{current_path}.{name}" if current_path else name
        elif access_type == 'item':
          current_full_path = f'{current_path}["{name}"]'
        else:
          continue

        pollutables[current_full_path] = self.summerize_value(value)

        # Decide whether to enqueue for further search
        type_info = get_type(value)
        continue_search_flag = (
          (current_layer < max_layer or max_layer == -1) and
          not is_c_wrapper(value) and
          not is_primitive(type_info)     
        )

        if continue_search_flag:
          logging.debug(f"Enqueuing attribute: {current_full_path} at layer {current_layer + 1}")
          queue.append((value, current_layer + 1, current_full_path))

    return pollutables

  def summerize_value(self, value):
    """
    @description
    ---------------------
    Generate the summary of the value.

    For class object, the summary is (class_name, class_name, standard?, writable?)
    For class itself, the summary is ("class", class_name, standard?, writable?)

    (type, name, standard?, writable?)

    @params value: The value to summarize.
    @return: The summary of the value.
    """
    type_info = get_type(value)
    name_info = get_name_info(value)
    is_standard = is_from_standard_library(value)
    writable = is_c_written(value, type_info)
    return (type_info, name_info, is_standard, writable)

  def parse_path(self, path):
    components = []
    i = 0
    n = len(path)
    while i < n:
      if path[i] == '.' and i+1 < n and path[i+1] == '[':
        i += 1
        if path[i] == '[':
          if i+1 < n and path[i+1] == '"':
            j = path.find('"]', i+2)
            if j == -1:
              raise ValueError(f"Invalid path: unterminated [\" at position {i}")
            name = path[i+2:j]
            components.append(('item', name))
            i = j + 2
          else:
            raise ValueError(f"Invalid item access at position {i}")
        else:
          raise ValueError(f"Unexpected character '{path[i]}' at position {i}")
      elif path[i] == '.':
        j = i + 1
        while j < n and (path[j].isalnum() or path[j] == '_'):
          j += 1
        attr = path[i+1:j]
        components.append(('attr', attr))
        i = j
      elif path[i] == '[':
        if i+1 < n and path[i+1] == '"':
          j = path.find('"]', i+2)
          if j == -1:
            raise ValueError(f"Invalid path: unterminated [\" at position {i}")
          name = path[i+2:j]
          components.append(('item', name))
          i = j + 2
        else:
          raise ValueError(f"Unsupported item access syntax at position {i}")
      else:
        j = i
        while j < n and (path[j].isalnum() or path[j] == '_'):
          j += 1
        attr = path[i:j]
        components.append(('attr', attr))
        i = j
    return components

  def find(self, path):
    """
    Find the object based on the path.

    @params path: The path to find the object.
    @return obj: The object found based on the path.
    """
    components = self.parse_path(path)
    obj = self.target
    for access_type, name in components:
      if access_type == 'attr':
        obj = getattr(obj, name)
      elif access_type == 'item':
        obj = obj[name]
      else:
        raise ValueError(f"Unknown access type: {access_type}")
    return obj
  
  def select(self, query):
    """
    Select the object based on the query.

    @params query: The query to select the object.
    Supported queries:
    - "type=module" -> Select the modules.
    - "type=callable" -> Select the callables.
    - "type=method" -> Select the methods.
    - "type=function" -> Select the functions.
    - "type=noncallable" -> Select the non-callable attributes.
    - "type=dict" -> Select the dictionaries.
    - "type=class" -> Select the classes.
    - "type=dataclass" -> Select the dataclasses.
    - "type=string" -> Select the strings.
    - "type=number" -> Select the numbers.
    - "type=boolean" -> Select the booleans.
    - "builtin=no" -> Select the user-defined objects.
    - "name=???" -> Select the object with __name__=???.
    - Combined queries like "type=class&builtin=no" -> Select user-defined classes.

    @return selected: A dictionary of selected paths and their values.
    """
    selected = {}
    query_pairs = [q.split('=') for q in query.split('&')]
    
    for path, value in self.summary.items():
      type_info, name_info, is_standard, writable = value
      match = True
  
      for key, q_value in query_pairs:
        if key == 'type':
          if not self._match_type(q_value, type_info, value):
            match = False
            break
        elif key == 'builtin':
          if not self._match_builtin(q_value, is_standard):
            match = False
            break
        elif key == 'name':
          if not self._match_name_precise(name_info, type_info, q_value):
            match = False
            break
        elif key == 'namelike':
          if not self._match_name(name_info, type_info, q_value):
            match = False
            break
        else:
          match = False
          break
  
      if match:
        selected[path] = value
    
    return selected

  def select_by_func(self, func):
    """
    Select the object based on the function.

    @params func: The function to select the object.
    @return selected: A dictionary of selected paths and their values.
    """
    selected = {}
    for path, value in self.summary.items():
      type_info, name_info, is_standard, writable = value
      if func(type_info, name_info, is_standard, writable):
        selected[path] = value
    return selected

  def _match_name_precise(self, name_info, type_info, match_name):
    """Handle name-based matching"""
    return match_name == name_info

  def _match_name(self, name_info, type_info, match_name):
    """Handle name-based matching"""
    return match_name in name_info

  def _match_type(self, q_value, type_info, full_value):
    """Handle type-based matching"""
    direct_types = {
      'module': 'module',
      'function': 'function',
      'OrderedDict': 'OrderedDict',
      'defaultdict': 'defaultdict',
      'deque': 'deque',
      'Counter': 'Counter',
      'ChainMap': 'ChainMap',
      'UserDict': 'UserDict',
      'UserList': 'UserList',
      'UserString': 'UserString',
      'class': 'class',
      'string': 'str',
      'boolean': 'bool'
    }
    
    if q_value in direct_types:
      return type_info == direct_types[q_value]
    
    if q_value == 'callable':
      return is_callable(type_info)
    if q_value == 'noncallable':
      return not is_callable(type_info)
    if q_value == 'method':
      return type_info in {"method", "builtin_function_or_method", "method-wrapper"}
    if q_value == 'number':
      return type_info in {"int", "float", "complex"}
    if q_value == 'dataclass':
      return hasattr(full_value[0], '__dataclass_fields__')
    if q_value == 'object':
      return is_instance(type_info)
    if q_value == 'dict':
      return support_getField_op(type_info) and type_info != 'str'
    return False

  def _match_builtin(self, q_value, is_standard):
    if q_value == 'yes':
      return is_standard
    if q_value == 'no':
      return not is_standard
    return False


#!/usr/bin/env python3

"""
@description
---------------------
This script contains the useful gadgets to find the pollutes the objects reflectively during runtime.
"""

def pollute(obj, layer=1, default_val="", method_only=False):
  """
  Pollute the object by changing its attributes or methods.

  @params obj: The object to pollute.
  @params layer: The current layer of the object.
  @params default_val: The default value to pollute the object with.
  @params method_only: If True, only pollute the methods of the object.
  """
  for attr_name in dir(obj):
    attr = getattr(obj, attr_name)
    if method_only and not callable(attr):
      continue
    try:
      setattr(obj, attr_name, default_val)
      print(f"Polluted {attr_name} to {default_val}")
    except Exception as e:
      print(f"Failed to pollute {attr_name} due to {e}")